<?php declare(strict_types=1);

namespace Everyware\Concepts;

use Everyware\Concepts\Contracts\ConceptProvider;
use Everyware\Concepts\Contracts\ConceptRepository;
use Everyware\Concepts\Events\ConceptApiEvent;
use Everyware\Concepts\Exceptions\ConceptDeleteError;
use Everyware\Concepts\Exceptions\ConceptCreateError;
use Everyware\Concepts\Exceptions\ConceptUpdateError;
use Everyware\Concepts\Exceptions\InvalidConceptData;
use Everyware\Concepts\Wordpress\Contracts\WpAction;
use Exception;
use Infomaker\Everyware\Support\NewRelicLog;

/**
 * Class AjaxController
 * @package Everyware\Concepts
 */
class ConceptController
{
    /**
     * @var ConceptRepository
     */
    private $repository;

    /**
     * @var ConceptProvider
     */
    private $provider;

    /**
     * @var array
     */
    private $responseCodes = [];

    /**
     * @var WpAction
     */
    private $wpAction;

    /**
     * @var LockHandler
     */
    private $lockHandler;

    /**
     * @var bool
     */
    private $logErrors;

    public function __construct(
        ConceptRepository $repository,
        ConceptProvider $provider,
        WpAction $wpAction,
        LockHandler $lockHandler,
        bool $logErrors = true
    )
    {
        $this->repository = $repository;
        $this->provider = $provider;
        $this->wpAction = $wpAction;
        $this->lockHandler = $lockHandler;
        $this->logErrors = $logErrors;
    }

    public function create($uuid): ConceptApiResponse
    {
        if ($this->conceptExists($uuid)) {
            return ConceptApiResponse::alreadyExists();
        }

        if ( ! $this->lockHandler->requestLock($uuid)) {

            // Do nothing if it is being created by another process.
            if ($this->conceptExists($uuid)) {
                return ConceptApiResponse::alreadyExists();
            }

            $e = new ConceptCreateError('Concept ' . $uuid . ' did not get created despite having been locked by another process.');
            $this->log('ConceptCreateError in concept CREATE: ', $e);

            return ConceptApiResponse::internalError($e);
        }

        $concept = null;

        try {
            try {
                $concept = $this->fetchConcept($uuid);
            } catch (InvalidConceptData $e) {
                $this->log('InvalidConceptData in concept CREATE: ', $e);

                return ConceptApiResponse::internalError($e);
            }

            if ( ! $concept instanceof OcConcept) {
                return ConceptApiResponse::notFoundInOC();
            }

            return $this->createPost($uuid, $concept);

        } finally {
            $this->lockHandler->unlock($uuid);
        }
    }

    /**
     * @param $uuid
     *
     * @return ConceptApiResponse
     */
    public function show($uuid): ConceptApiResponse
    {
        $post = $this->repository->findByUuid($uuid);

        if ( ! $post instanceof ConceptPost) {
            return ConceptApiResponse::error([], 404)
                ->addResponseCode(ConceptApiResponse::NOT_FOUND_IN_WP);
        }

        return ConceptApiResponse::success($this->extractPostData($post, $uuid));
    }

    /**
     * @param $uuid
     *
     * @return ConceptApiResponse
     */
    public function update($uuid): ConceptApiResponse
    {
        $post = $this->repository->findByUuid($uuid);

        if ( ! $post instanceof ConceptPost) {
            return ConceptApiResponse::notFoundInDb();
        }

        $concept = null;

        try {
            $concept = $this->fetchConcept($uuid);
        } catch (InvalidConceptData $e) {
            $this->log('InvalidConceptData in concept UPDATE: ', $e);

            return ConceptApiResponse::internalError($e);
        }

        if ( ! $concept instanceof OcConcept) {
            return ConceptApiResponse::notFoundInOC();
        }

        return $this->updatePost($uuid, $post, $concept);
    }

    public function remove($uuid): ConceptApiResponse
    {
        $post = $this->repository->findByUuid($uuid);

        if ( ! $post instanceof ConceptPost) {
            return ConceptApiResponse::notFoundInDb();
        }

        return $this->removePost($uuid, $post);
    }

    public function synchronize($uuid): ConceptApiResponse
    {
        $post = $this->repository->findByUuid($uuid);
        $concept = null;

        try {
            $concept = $this->fetchConcept($uuid);
        } catch (InvalidConceptData $e) {
            $this->log('InvalidConceptData in concept UPDATE: ', $e);

            return ConceptApiResponse::internalError($e);
        }

        if ( ! $concept instanceof OcConcept) {
            // Remove concept that can't be retrieved from Open Content but was found in Wordpress
            return $post instanceof ConceptPost ? $this->removePost($uuid, $post) : ConceptApiResponse::notFoundInOC();
        }

        if ( ! $this->lockHandler->requestLock($uuid)) {

            // Do nothing if it is being created by another process.
            if ($this->conceptExists($uuid)) {
                return ConceptApiResponse::alreadyExists();
            }

            $e = new ConceptCreateError('Concept ' . $uuid . ' did not get created despite having been locked by another process.');
            $this->log('ConceptCreateError in concept SYNCHRONIZE: ', $e);

            return ConceptApiResponse::internalError($e);
        }

        try {
            // Create concepts that can't be found in Wordpress.
            if ( ! $post instanceof ConceptPost) {
                $response = $this->createPost($uuid, $concept);

                return $response;
            }

            return $this->updatePost($uuid, $post, $concept);

        } finally {
            $this->lockHandler->unlock($uuid);
        }
    }

    private function createPost($uuid, OcConcept $concept): ConceptApiResponse
    {
        try {

            $this->verifyParentExists($concept);

            $newPost = $this->repository->create($concept);

            $this->triggerEvent('created', new ConceptApiEvent($uuid, $newPost));

        } catch (ConceptCreateError $e) {
            $this->log('ConceptCreateError in concept CREATE: ', $e);

            return ConceptApiResponse::internalError($e);
        } finally {
            $this->lockHandler->unlock($uuid);
        }

        return ConceptApiResponse::creationSuccess($this->getResponseCodes());
    }

    private function removePost(string $uuid, ConceptPost $post): ConceptApiResponse
    {
        try {
            $this->repository->delete($post);

            $this->triggerEvent('deleted', new ConceptApiEvent($uuid, $post));

        } catch (ConceptDeleteError $e) {
            $this->log('ConceptDeleteError in concept DELETE: ', $e);

            return ConceptApiResponse::internalError($e);
        }

        return ConceptApiResponse::deletionSuccess($this->getResponseCodes());
    }

    private function updatePost($uuid, ConceptPost $post, OcConcept $concept): ConceptApiResponse
    {
        try {
            $this->verifyParentExists($concept);

            $newPost = $this->repository->updateByOcConcept($concept);

            if ($post->post_parent !== $newPost->post_parent) {
                $this->addResponseCode(ConceptApiResponse::MOVED);
            }

            $this->triggerEvent('updated', new ConceptApiEvent($uuid, $newPost));

        } catch (ConceptUpdateError $e) {
            $this->log('ConceptUpdateError in concept UPDATE: ', $e);

            return ConceptApiResponse::internalError($e);
        }

        return ConceptApiResponse::updateSuccess($this->getResponseCodes());
    }

    /**
     * @param string $uuid
     *
     * @return OcConcept
     * @throws InvalidConceptData
     */
    private function fetchConcept($uuid): ?OcConcept
    {
        return $this->provider->getSingle($uuid, OcConcept::requiredProperties(1));
    }

    /**
     * @param ConceptPost $post
     * @param string|null $uuid
     *
     * @return array
     */
    private function extractPostData(ConceptPost $post, $uuid = null): array
    {
        $parent = $post->parent();
        $postId = $post->id;

        return [
            'uuid' => $uuid ?? $this->repository->findUuidById($postId),
            'postId' => $postId,
            'permalink' => $post->permalink(),
            'parent' => $parent instanceof ConceptPost ? $this->extractPostData($parent) : null
        ];
    }

    /**
     * Determine if a concept already exists in Wordpress
     *
     * @param $uuid
     *
     * @return bool
     */
    private function conceptExists($uuid): bool
    {
        $this->lockHandler->waitWhileLocked($uuid);

        return $this->repository->findByUuid($uuid) instanceof ConceptPost;
    }

    private function triggerEvent($eventName, ConceptApiEvent $event): void
    {
        $this->wpAction->doAction("ew_concept_{$eventName}", $event);
    }

    private function verifyParentExists(OcConcept $concept): bool
    {
        $parentUuid = $concept->getParentUuid();

        // Do nothing if no parent uuid can be found on concept or if parent already exists in database
        if ( ! $parentUuid || $this->conceptExists($parentUuid)) {
            return true;
        }

        if ( ! $this->lockHandler->requestLock($parentUuid)) {

            // Do nothing if it is being created by another process.
            if ($this->conceptExists($parentUuid)) {
                return true;
            }

            $this->addResponseCode(ConceptApiResponse::PARENT_NOT_CREATED);

            $e = new ConceptCreateError('Concept ' . $parentUuid . ' did not get created despite having been locked by another process.');
            $this->log('ConceptCreateError in concept VERIFYPARENTEXISTS: ', $e);

            return false;
        }

        $this->addResponseCode(ConceptApiResponse::PARENT_NOT_FOUND_IN_WP);

        $parent = $this->fetchParent($concept);

        try {
            // If none is retrieved we set the response code for it.
            if ( ! $parent instanceof OcConcept) {
                $this->addResponseCode(ConceptApiResponse::PARENT_NOT_FOUND_IN_SOURCE);

                return false;
            }

            $response = $this->createPost($parentUuid, $parent);
            if ($response->getStatusCode() !== 201) {
                $this->addResponseCode(ConceptApiResponse::PARENT_NOT_CREATED);

                return false;
            }

            $this->addResponseCode(ConceptApiResponse::PARENT_CREATED);

            return true;

        } finally {
            $this->lockHandler->unlock($parentUuid);
        }
    }

    private function fetchParent(OcConcept $concept): ?OcConcept
    {
        // Has the parent been fetched as a relation? or request OC for the parent
        try {
            return $concept->getParent() ?? $this->fetchConcept($concept->getParentUuid());
        } catch (InvalidConceptData $e) {
            $this->addResponseCode(ConceptApiResponse::INTERNAL_ERROR);
        }

        return null;

    }

    private function addResponseCode(string $code): void
    {
        $this->responseCodes[] = $code;
    }

    /**
     * @return array
     */
    private function getResponseCodes(): array
    {
        return $this->responseCodes;
    }

    private function log(string $message, Exception $e): void
    {
        if ($this->logErrors) {
            NewRelicLog::error($message, $e);
        }
    }
}
